
# This file is part of FatxFast.
#
#    FatxFast is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    FatxFast is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with FatxFast.  If not, see <http://www.gnu.org/licenses/>.

# Versioning based on: 
# http://en.wikipedia.org/wiki/Versioning#Designating_development_stage
__author__ = "dryatu (c) 2013"
__version__ = "1.1.2"


from FatxFast.constants import *
import pygame
from FatxFast.objects import entity
import time

class MoveableEntity(entity.GameEntity):
    """Base class for objects which are moveable on the map. In addition
    to normal GameEntity it has:
    Attributes:
        Speedvector(Vec2), AccelVector(Vec2), Max & Min speeds(Floats), 
        Collides(Boolean)
    Methods:
        Move by speed: Does collision detection based on speed
        moves entity next to the colliding object if one is present.

    Collision detection in detail:
        Fetches the border area tile positions for the entity ->
        Fetches the end position for the axis -> Does a tile by tile
        iteration from starting position to the end position. Upon collision
        returns the colliding tile -> Entity is moved to the new position.
        If there was a collision the new maximum speed is so that next move
        will move the entity next to the colliding point.
    """
    def __init__(self, gamescene, **kwargs):
        self.gamescene = gamescene
        super(MoveableEntity, self).__init__(**kwargs)
        self.speedvector = pygame.math.Vector2(0,0)
        self.collides = kwargs.get("collides", True)
        self.maxspeed = CHAR_MAX_SPEED
        self.minspeed = CHAR_MIN_SPEED
        self._slow_down_factor = 2
        self.accelspeed = CHAR_ACCEL
        self.accelvector = pygame.math.Vector2(0,0)
        self._collision_sprite = pygame.sprite.Sprite()
    
    def move(self, x=0, y=0):
        pos = self.get_position()
        self.set_position([pos[0]+x, pos[1]+y])
    
    def get_speed(self):
        return self.speedvector

    def set_speed(self, x=None, y=None):
        if x is not None:
            self.speedvector.x = x
        if y is not None:
            self.speedvector.y = y

    def get_accel(self):
        return self.accelvector

    def set_accel(self, x=None, y=None):
        if x is not None:
            self.accelvector.x = x
        if y is not None:
            self.accelvector.y = y

    def _get_collision_tile(self, x, y):
        return self.gamescene.tilemap.sprite_layers[
            COLLIDE_LAYER].content2D[y][x]

    def _get_first_obstacle(self, xstart, xend, ystart, yend, 
        fstep=0, sstep=0):
        for first in xrange(ystart, yend+fstep, fstep):
            for second in xrange(xstart, xend+sstep, sstep):
                collision = self._get_collision_tile(second,first)
                if collision is not None:
                    return collision

    def _move_on_x_axis(self):
        """Move entity by x speed and check for collisions on x axis. 
        
        There's a collision => speedvector has to be resized to the colliding
        edge of the tile. Create a collision rect => trigger events. If there's
        a place changing trigger => Use a linear interpretation
        of the path triggers i.e. speedvector.x > 0 => trigger events from
        current_pos_x -> place_change_trigger"""
        # Defines border area tile positions
        y = self.gamescene.tilemap.get_tile_ypos(self.rect.top)
        ey = self.gamescene.tilemap.get_tile_ypos(self.rect.bottom-1)
        x = self.gamescene.tilemap.get_tile_xpos(self.rect.x)
        if self.collides:
            if self.speedvector.x < 0:
                ex = self.gamescene.tilemap.get_tile_xpos(
                    self.rect.x+self.speedvector.x)
                col = self._get_first_obstacle(
                    x, ex, y, ey, fstep=1, sstep=-1)
                if col:
                    self.speedvector.x = col.rect.right-self.rect.left
            elif self.speedvector.x > 0:
                ex = self.gamescene.tilemap.get_tile_xpos(
                    self.rect.right-1+self.speedvector.x)
                col = self._get_first_obstacle(
                    x, ex, y, ey, fstep=1, sstep=1)
                if col:
                    self.speedvector.x = col.rect.left-self.rect.right
            self._set_collision_rect("x", self.speedvector.x)
            self._trigger_events()
        self.move(self.speedvector.x, 0)

    def _move_on_y_axis(self):
        """Move entity by y speed and check for collisions on y axis."""
        x = self.gamescene.tilemap.get_tile_xpos(self.rect.left)
        ex = self.gamescene.tilemap.get_tile_xpos(self.rect.right-1)
        y = self.gamescene.tilemap.get_tile_ypos(self.rect.y)
        if self.collides: 
            if self.speedvector.y < 0:
                ey = self.gamescene.tilemap.get_tile_ypos(
                    self.rect.y+self.speedvector.y)
                col = self._get_first_obstacle(x, ex, y, ey, fstep=-1, sstep=1)
                if col:
                    self.speedvector.y = col.rect.bottom-self.rect.top
            elif self.speedvector.y > 0:
                ey = self.gamescene.tilemap.get_tile_ypos(
                    self.rect.bottom-1+self.speedvector.y)
                col = self._get_first_obstacle(x, ex, y, ey, fstep=1, sstep=1)
                if col:
                    self.speedvector.y = col.rect.top-self.rect.bottom
            self._set_collision_rect("y", self.speedvector.y)
            self._trigger_events()
        self.move(0, self.speedvector.y)

    def _move_by_speed(self, dt):
        self._slow_down('x')
        self._slow_down('y')
        self.speedvector = self.speedvector+(self.accelvector*dt/8)
        if self.speedvector.length() > self.maxspeed:
            self.speedvector.scale_to_length(self.maxspeed)  

        # Check for collisions on both axes & move by speed
        self._move_on_x_axis()
        self._move_on_y_axis()

    def _set_collision_rect(self, component, size):
        # Sets a rectangle from the speedvector component
        rect = self.rect.copy()
        if component == "x":
            size_component = "width"
        elif component == "y":
            size_component = "height"
        else:
            return False
        if size < 0:
            component_value = getattr(rect, component)
            setattr(rect, component, component_value+size)
        size_value = getattr(rect, size_component)
        setattr(rect, size_component, size_value+abs(size))
        self._collision_sprite.rect = rect

    def _slow_down(self, axis):
        self._stop_movement(axis)
        if not getattr(self.accelvector, axis):
            if getattr(self.speedvector, axis):
                setattr(self.accelvector, axis, 
                    -getattr(self.speedvector.normalize(), axis)*\
                        self._slow_down_factor)

    def _stop_movement(self, axis):
        if (int(abs(getattr(self.speedvector, axis))/self._slow_down_factor)\
            < self.minspeed):
            setattr(self.speedvector, axis, 0)

    def _trigger_events(self):
        pass

    def update(self, dt):
        self._move_by_speed(dt)

